// lite-c newton demo (c) 2007 thomas oppl



#define NEWTON_NULL 0
#define NEWTON_SPHERE 1
#define NEWTON_BOX 2
#define NEWTON_CONE 3
#define NEWTON_CAPSULE 4
#define NEWTON_CYLINDER 5
#define NEWTON_CHAMFEREDCYLINDER 6
#define NEWTON_CONVEXHULL 7

NewtonWorld* nworld;
var newton_running = 0;

typedef struct
{
	float x, y, z;
	float nx, ny, nz;
	float tu1, tv1;
	float tu2, tv2;
	float tx3, ty3, tz3, tw3;
} _VERTEX_;



//#######################################################################################
//####################################################################################### TEMPORARY FLOAT VECTORS
//#######################################################################################

// rotating use of a temporary float vector
float _fvector[192]; // 3 * 64

//--------------------------------------------------------------------------------------- vectorf()
// vectorf() works like vector() but returns a pointer to 3 floats instead of 3 vars
//---------------------------------------------------------------------------------------
float *vectorf(float x, float y, float z)
{
	static int i = 0;
	i += 3;
	if(i >= 192) i = 0;
	_fvector[i+0] = x; _fvector[i+1] = y; _fvector[i+2] = z;
	return &_fvector[i];
}

//--------------------------------------------------------------------------------------- vectorf()
// overloaded version which can convert a var vector to a float vector
//---------------------------------------------------------------------------------------
float *vectorf(VECTOR* v)
{
	return vectorf(v->x, v->y, v->z);
}



//#######################################################################################
//####################################################################################### STATIC COLLISION GEOMETRY
//#######################################################################################

//--------------------------------------------------------------------------------------- newton_treecollisionaddentity()
// add the triangles of the entity to the static collision geometry
//
// (note: if you would like to add WED blocks as static collision geometry then you have
// to pass NULL to ent_getmesh and loop through all blocks! the transformation to world
// space isn't necessary then.) 
//---------------------------------------------------------------------------------------
void newton_treecollisionaddentity(NewtonCollision* treecollision, ENTITY* entity)
{
	if(!entity) return;
	
	LPD3DXMESH pmesh = (LPD3DXMESH)ent_getmesh(entity, 0, 0);
	int numvertices = pmesh->GetNumVertices();
	int numfaces = pmesh->GetNumFaces();
	
	_VERTEX_ *pvertices; pmesh->LockVertexBuffer(0, (void**)&pvertices);
	short *pindices; pmesh->LockIndexBuffer(0, (void**)&pindices);
	int *pattributes; pmesh->LockAttributeBuffer(0, &pattributes);
	
	// transform vertices to world space
	D3DXVECTOR4 *ptransformedvertices = (D3DXVECTOR4*)malloc(sizeof(D3DXVECTOR4) * numvertices);
	D3DXMATRIX m; ent_getmatrix(entity, &m);
	D3DXVECTOR3 tempvector;
	int i;
	for(i = 0; i < numvertices; i++)
	{
		tempvector.x = pvertices[i].x;
		tempvector.y = pvertices[i].z; // direct3d -> 3dgs coordinate system
		tempvector.z = pvertices[i].y; // direct3d -> 3dgs coordinate system 
		D3DXVec3Transform(&ptransformedvertices[i], &tempvector, &m);	
	}
	
	// add triangles to collision tree
	for(i = 0; i < numfaces; i++)
	{	
		float v[9];
		v[0] = ptransformedvertices[pindices[(i*3)+2]].x * QUANTTOMETER;
		v[1] = ptransformedvertices[pindices[(i*3)+2]].y * QUANTTOMETER;
		v[2] = ptransformedvertices[pindices[(i*3)+2]].z * QUANTTOMETER;
		v[3] = ptransformedvertices[pindices[(i*3)+1]].x * QUANTTOMETER;
		v[4] = ptransformedvertices[pindices[(i*3)+1]].y * QUANTTOMETER;
		v[5] = ptransformedvertices[pindices[(i*3)+1]].z * QUANTTOMETER;
		v[6] = ptransformedvertices[pindices[(i*3)+0]].x * QUANTTOMETER;
		v[7] = ptransformedvertices[pindices[(i*3)+0]].y * QUANTTOMETER;
		v[8] = ptransformedvertices[pindices[(i*3)+0]].z * QUANTTOMETER;
		NewtonTreeCollisionAddFace(treecollision, 3, v, 12, pattributes[i]);
	}
	
	free(ptransformedvertices);
	
	pmesh->UnlockVertexBuffer();
	pmesh->UnlockIndexBuffer();
	pmesh->UnlockAttributeBuffer();
}

//--------------------------------------------------------------------------------------- newton_addstaticcollisiongeometry()
// add entities with flag8 to the static collision geometry
//---------------------------------------------------------------------------------------
void newton_addstaticcollisiongeometry()
{
	NewtonCollision* treecollision = NewtonCreateTreeCollision(nworld, 0);
	NewtonTreeCollisionBeginBuild(treecollision);
	you = ent_next(0);
 	while(you != 0)
	{
		if(you->flags & FLAG8)
		{
			newton_treecollisionaddentity(treecollision, you);
 		}
 		you = ent_next(you);
 	}
 	NewtonTreeCollisionEndBuild(treecollision, 0);
	NewtonBody* level = NewtonCreateBody(nworld, treecollision);
	NewtonReleaseCollision(nworld, treecollision);
	
	NewtonSetWorldSize(nworld, vectorf(-1000, -1000, -1000), vectorf(1000, 1000, 1000)); // todo: make size dependent on tree collision bounding box
}



//#######################################################################################
//####################################################################################### NEWTON MAIN LOOP
//#######################################################################################

//--------------------------------------------------------------------------------------- newton_start()
// create the newton world and start the newton main loop
//--------------------------------------------------------------------------------------- 
void newton_start()
{
	newton_running = 1;
	nworld = NewtonCreate(0, 0);
	
	newton_addstaticcollisiongeometry();
	
	while(newton_running) // newton has to update every frame
	{
		NewtonUpdate(nworld, time_step / 16);
		wait(1);
	}
}

//--------------------------------------------------------------------------------------- newton_stop()
// stop the newton main loop and destroy the newton world
//--------------------------------------------------------------------------------------- 
void newton_stop()
{
	newton_running = 0;
	NewtonDestroy(nworld);
}



//#######################################################################################
//####################################################################################### REGISTER PHYSICS ENTITY
//#######################################################################################

//--------------------------------------------------------------------------------------- onsettransform()
// callback which updates the transformation of entities registered with newton
//---------------------------------------------------------------------------------------
void onsettransform(NewtonBody *body, float *m)
{
	ENTITY *entity = (ENTITY*)NewtonBodyGetUserData(body);
	ent_setmatrix_rb(entity, m);
}

//--------------------------------------------------------------------------------------- newton_addentity()
// register the entity with newton
//---------------------------------------------------------------------------------------
NewtonBody *newton_addentity(ENTITY* entity, float mass, int collisiontype, void* forceandtorquecallback)
{
	if(!entity) return;
	
	NewtonCollision* collision;
	
	VECTOR vmin; vec_for_min(&vmin, entity);
	VECTOR vmax; vec_for_max(&vmax, entity);
	vmin.x *= entity->scale_x; vmin.y *= entity->scale_y; vmin.z *= entity->scale_z;
	vmax.x *= entity->scale_x; vmax.y *= entity->scale_y; vmax.z *= entity->scale_z;
	
	D3DXMATRIX moffset; // offset from center of bounding box to coordinate origin of entity
	D3DXMatrixTranslation(&moffset,  (vmin.x + (vmax.x - vmin.x) / 2.0) * QUANTTOMETER,
												(vmin.y + (vmax.y - vmin.y) / 2.0) * QUANTTOMETER,
												(vmin.z + (vmax.z - vmin.z) / 2.0) * QUANTTOMETER);
	
	switch(collisiontype)
	{
		//---------------------------------------------------
		case NEWTON_SPHERE:
			
			float radiusx = (vmax.x - vmin.x) / 2.0;
			float radiusy = (vmax.y - vmin.y) / 2.0;
			float radiusz = (vmax.z - vmin.z) / 2.0;
			
			collision = NewtonCreateSphere(nworld,	radiusx * QUANTTOMETER,
																radiusy * QUANTTOMETER,
																radiusz * QUANTTOMETER, &moffset);
			
			break;
		
		//---------------------------------------------------	
		case NEWTON_BOX:
			
			float sizex = (vmax.x - vmin.x);
			float sizey = (vmax.y - vmin.y);
			float sizez = (vmax.z - vmin.z);
			
			collision = NewtonCreateBox(nworld, sizex * QUANTTOMETER,
															sizey * QUANTTOMETER,
															sizez * QUANTTOMETER, &moffset);
			
			break;
		
		//---------------------------------------------------	
		case NEWTON_CONE:
			
			float radius = (vmax.z - vmin.z) / 2;
			float height = (vmax.x - vmin.x);
			
			collision = NewtonCreateCone(nworld, radius * QUANTTOMETER, height * QUANTTOMETER, &moffset);
			
			break;
		
		//---------------------------------------------------	
		case NEWTON_CAPSULE:
			
			float radius = (vmax.z - vmin.z) / 2;
			float height = (vmax.x - vmin.x);
			
			collision = NewtonCreateCapsule(nworld, radius * QUANTTOMETER, height * QUANTTOMETER, &moffset);
			
			break;
		
		//---------------------------------------------------	
		case NEWTON_CYLINDER:
			
			float radius = (vmax.z - vmin.z) / 2;
			float height = (vmax.x - vmin.x);
			
			collision = NewtonCreateCylinder(nworld, radius * QUANTTOMETER, height * QUANTTOMETER, &moffset);
			
			break;
		
		//---------------------------------------------------	
		case NEWTON_CHAMFEREDCYLINDER:
			
			float radius = (vmax.z - vmin.z) / 2;
			float height = (vmax.x - vmin.x);
			
			collision = NewtonCreateChamferCylinder(nworld, radius * QUANTTOMETER, height * QUANTTOMETER, &moffset);
			
			break;
		
		//---------------------------------------------------	
		case NEWTON_CONVEXHULL:
			
		   int vertices = ent_vertices(entity);
			float *vertexcloud = (float*)malloc(sizeof(float) * vertices * 3);
			
			int i;
			for(i = 0; i < vertices; i++)
			{
				VECTOR vertex; vec_for_mesh(&vertex, entity, i + 1);
				vertexcloud[i * 3 + 0] = vertex.x * entity->scale_x * QUANTTOMETER;
				vertexcloud[i * 3 + 1] = vertex.y * entity->scale_y * QUANTTOMETER;
				vertexcloud[i * 3 + 2] = vertex.z * entity->scale_z * QUANTTOMETER;
			}
			
			collision = NewtonCreateConvexHull(nworld, vertices, vertexcloud, 12, NULL);
			free(vertexcloud);
			
			break;
	}
	
	NewtonBody* body = NewtonCreateBody(nworld, collision);
	float m[16]; ent_getmatrix_rb(entity, m);
	NewtonBodySetMatrix(body, m);
	
	entity->skill99 = body; // save body pointer with entity
	NewtonBodySetUserData(body, entity); // save entity pointer with body
	
	NewtonBodySetTransformCallback(body, onsettransform);
	NewtonBodySetForceAndTorqueCallback(body, forceandtorquecallback);
   
   float inertia[3];
   float centerofmass[3];
   NewtonConvexCollisionCalculateInertialMatrix(collision, inertia, centerofmass); 
	NewtonBodySetMassMatrix(body, mass, mass * inertia[0], mass * inertia[1], mass * inertia[2]);
   
   NewtonReleaseCollision(nworld, collision);
   
   return body;
}